/**
 * @file   utils.hh
 * @author Anthony Truchet <anthony.truchet@logilab.fr>
 * @date   Tue Oct 30 11:23:41 2012
 *
 * @brief  Misc utilities used and provided by Yams++
 *
 *
 */

#ifndef YAMS__UTILS_H__INCLUDED
#define YAMS__UTILS_H__INCLUDED

#include <iterator>
#include <algorithm>

#include <boost/config.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
// We are not on a recent enough compiler and use the boost version
// namespace yams {

typedef size_t entity_id;
// On a C++'11 or TR1 compliant compiler we should instead use :
/* #include <cstdint>
 * typedef intptr_t entity_id;
 */

// }

/**
 * @brief Input sequences after Stroustrup 13.3.1 and convenience helpers
 * XXX could/should (?) be replaced by (or built upon) boost::range
 */
//@{

template <class In, class In_const = In>
struct ConvenientIterable :
        std::pair<In, In>
{
    typedef In iterator;
    typedef In_const const_iterator; // Type required by BOOST_FOREACH
    typedef ConvenientIterable<iterator, const_iterator> self_type;
    typedef std::pair<iterator, iterator> base_type;

    // Required by the iterator concept
    typedef typename iterator::iterator_category  iterator_category;
    typedef typename iterator::value_type         value_type;
    typedef typename iterator::difference_type    difference_type;
    typedef typename iterator::pointer            pointer;
    typedef typename iterator::reference          reference;

    // Constructors taking care of implicit conversions and const-ness
    template<class In1, class In2>
    ConvenientIterable(In1 i1, In2 i2) : base_type(i1, i2) {}

    template<class In1, class In2>
    ConvenientIterable(std::pair<In1, In2>& p) : base_type(p) {}

    template<class In1, class In2>
    ConvenientIterable(const ConvenientIterable<In1, In2>& other) :
        base_type(other) {}

    // Commodity accessor required by BOOST_FOREACH
    iterator begin() const {return this->first;}
    iterator end()   const {return this->second;}

    // operator== and the like should be inherited from std::pair

    /**
     * @brief the bool cast test if further iteration is possible
     */
    operator bool() const { return this->first != this->second; }

    /**
     * @brief Just forward the dereference to the \c first member
     * @return
     */
    typename In::reference operator*()  { return *(this->first); }
    // TODO Provide const and arrow variants
    //typename In::pointer   operator->() { return this->first.operator->(); }

    /**
     * @brief prefix incrementation operator
     * @return reference to the incremented ConvenientIterator
     */
    self_type& operator++() { ++(this->first); return *this; }

    /**
     * @brief postfix incrementation operator
     * @return acopy of the ConvenientIterator before its incrementation
     */
    self_type operator++(int)
    {
        self_type& self = *this;
        self_type r(self);
        ++self;
        return r;
    }

    // NB: We do NOT provide decrementation operators on purpose

    /**
     * @defgroup container_short_hands Short hand for STL algorithms
     * Usual STL operation and algorithms on the underlying sequence
     */
    //@{

    inline iterator
    find(const value_type& looked_for) const
    { return std::find(this->first, this->second, looked_for); }

    template <typename Op>
    inline Op
    for_each(Op op) const
    { return std::for_each(this->first, this->second, op); }

    //XXX untested yet
    template <class Pred>
    inline iterator
    find_if(Pred pred) const
    { return std::find_if(this->first, this->second, pred); }

    //@}
};


/**
 * @brief Helper to build a convenient iterable from a STL container
 * @param c Reference to the container
 * @tparam C class of the contianer
 * @return a new convenient iterable
 */
template<class C>
ConvenientIterable<typename C::iterator, typename C::const_iterator>
iter(C& c)
{ return ConvenientIterable<typename C::iterator, typename C::const_iterator>(c.begin(), c.end()); }

/**
 * \brief Helper to build a convenient const iterable from a const STL container
 * @param c Reference to the (potentially const) container
 * @tparam C class of the contianer
 * @return a new convenient const iterable
 */
template<class C>
ConvenientIterable<typename C::const_iterator>
iter(const C& c)
{ return ConvenientIterable<typename C::const_iterator>(c.begin(), c.end()); }

//@}


#endif // YAMS__UTILS_H__INCLUDED
