/**
 * \file yams.hh
 * \brief Base classes for defining schemata
 *
 *  Created on: 6 nov. 2012
 *      Author: Anthony Truchet <anthony.truchet@logilab.fr>
 */

#ifndef YAMS__YAMS_H__INCLUDED
#define YAMS__YAMS_H__INCLUDED

#include <cassert>
#include <memory>

#include <boost/bimap/bimap.hpp>
//#include <boost/bimap/list_of.hpp>
#include <boost/bimap/unordered_set_of.hpp>
#include <boost/bimap/unordered_multiset_of.hpp>

// Boost bimaps namespace alias.
namespace bm = boost::bimaps;

#include <boost/range/adaptor/map.hpp>

// Boost::Serialize archive support
// #include <boost/archive/text_oarchive.hpp>
// #include <boost/archive/text_iarchive.hpp>

#include "utils.hh"

// namespace yams {

/**
 * @brief The base of all entity classes.
 *
 * This class is \em required to be \em virtually inherited by all
 * entity classes through the \c Entity template. It provide unique
 * integer identifier for each entity and some coinvenient dynamic
 * access and cast method.
 */
class AnyEntity
{
public:
    entity_id id() const
    { return reinterpret_cast<entity_id>(static_cast<const AnyEntity*>(this)); }

    virtual boost::shared_ptr<AnyEntity> ptr_any() = 0;
    virtual boost::shared_ptr<AnyEntity const> ptr_any() const = 0;

    template <class T>
    boost::shared_ptr<T> ptr()
    { return boost::dynamic_pointer_cast<T>(ptr_any()); }

    template <class T>
    boost::shared_ptr<T const> ptr() const
    { return boost::dynamic_pointer_cast<T const>(ptr_any()); }
};

/**
 * @brief The CRTP for defining class representing an Entity
 *
 * The intended use (namely CRTP) is \code
class MyEntityClass : public Entity<MyEntityClass>
{
    // The attributes and behaviour.
};
\endcode
 *
 * \b NB: This class must \em not be inherited virtually !
 *
 * For more information regarding the notion of CRTP, please refer
 * to a good C++ book or to
 * http://en.wikipedia.org/wiki/Curiously_recurring_template_pattern
 */
template <class DerivedT>
class Entity :
    public virtual AnyEntity,
    public boost::enable_shared_from_this<DerivedT>
{
public:
    typedef boost::shared_ptr<DerivedT> pointer;
    typedef boost::shared_ptr<const DerivedT> const_pointer;


    /**
     * @brief  Statically build a shared pointer to itself
     * @return a \c boost::shared_ptr to the object itself
     */
    pointer ptr()
    { return this->shared_from_this(); }

    /**
     * @brief  Statically build a shared pointer to itself, when const
     * @return a \c boost::shared_ptr to the const object itself
     */
    const_pointer ptr() const
    { return  this->shared_from_this(); }

    virtual boost::shared_ptr<AnyEntity> ptr_any()
    { return boost::static_pointer_cast<AnyEntity>(ptr()); }

    virtual boost::shared_ptr<AnyEntity const> ptr_any() const
    { return boost::static_pointer_cast<AnyEntity const>(ptr()); }

    /**
     * @brief Allow to iterate over the objects of a relation
     *
     * Allow to iter over the (shared_ptr to) C++ objects \c o such
     * that the \c Relation hold between \c *this as subject and \c o as
     * object.
     *
     * @return an implementation of the \c boost::range concept
     */
    template <class Relation>
    boost::select_second_const_range<typename Relation::subject_view::range_type>
    iter() // There may be a (big) const-ness problem (araising from) here
    { return Relation::for_subject(ptr()); }

    /**
      * @brief Allow to iterate over the subject of a relation
      *
      * Allow to iter over the (shared_ptr to) C++ objects \c s such
      * that the \c Relation hold between \c s as subject and \c *this
      * as object.
      *
      * @return an implementation of the \c boost::range concept
      */
    template <class Relation>
    boost::select_second_const_range<typename Relation::object_view::range_type>
    rev_iter() // There may be a (big) const-ness problem (araising from) here
    { return Relation::for_object(ptr()); }

    /**
     * @brief Add a relation
     *
     *  Add the \c Relation between \c *this as subject and \c obj
     *  as object. The relation should be garanted to exist after the
     *  call but the addition will not be considered successful if the
     *  relatation already existed before the call.
     *
     * @param obj the object of the relation to add
     *
     * @tparam Relation the type representing the RelationDefinition to add
     *
     * @return whether the add was successful
     */
    template <class Relation>
    bool add(typename Relation::object_pointer obj)
    {
        return Relation::add( ptr(), obj);
    }

    /**
     * @brief Remove a relation
     *
     *  Remove the \c Relation between \c *this as subject and \c obj
     *  as object. The relation should be garanted not to exist after the
     *  call but the removaal will not be considered successful if the
     *  relatation did not exist before the call.
     *
     * @param obj the object of the relation to remove
     *
     * @tparam Relation the type representing the RelationDefinition to remove
     *
     * @return whether the remove was successful
     */
    template <class Relation>
    bool remove(typename Relation::object_pointer obj)
    {
        return Relation::remove( ptr(), obj);
    }

    /**
     * @brief Test a relation
     *
     *  Test the \c Relation between \c *this as subject and \c obj
     *  as object.
     *
     * @param obj the object of the relation to test
     *
     * @tparam Relation the type representing the RelationDefinition to test
     *
     * @return whether \c *this is subject of \c obj for \c Relation
     */
    template <class Relation>
    bool rel(typename Relation::object_pointer obj)
    {
        return Relation::test( ptr(), obj);
    }

};

struct subject {};
struct object  {};

template<class RelationTypeT, class SubjectT, class ObjectT>
class RelationDefinition
{

public:
    typedef boost::shared_ptr< SubjectT > subject_pointer;
    typedef boost::shared_ptr< ObjectT >  object_pointer;

    // This is for '**' cardinality.
    // To implement '1*' use (unordered_)set_of
    typedef bm::bimap <
        bm::unordered_multiset_of<
            bm::tagged<subject_pointer, subject> >,
        bm::unordered_multiset_of<
            bm::tagged<object_pointer, object > >,
        bm::unordered_set_of_relation<>
        > bimap_type;

    typedef typename bimap_type::value_type relation;
    typedef typename bimap_type::template map_by<subject> subject_view;
    typedef typename bimap_type::template map_by<object>  object_view;


protected:
    static bimap_type& Instance()
    {
        // static std::auto_ptr<bimap_type> the_bimap(new bimap_type());
        static bimap_type the_bimap;
        return the_bimap;
    }

public:
    static bool add(subject_pointer subject, object_pointer object)
    {
        // insert returns a pair (insertion_pos : iterator, success : bool)
        return Instance().insert(relation(subject, object)).second;
    }

    static bool remove(subject_pointer subject, object_pointer object)
    {
        // erase return the number of elements
        return Instance().erase(relation(subject, object));
    }

    static bool test(subject_pointer subject, object_pointer object)
    {
        return Instance().find(relation(subject, object)) != Instance().end();
    }

    static typename boost::select_second_const_range<typename subject_view::range_type>
    for_subject(subject_pointer subject)
    {
        return Instance().left.equal_range(subject) | boost::adaptors::map_values;
    }

    static boost::select_second_const_range<typename object_view::range_type>
    for_object(object_pointer object)
    {
        return Instance().right.equal_range(object) | boost::adaptors::map_values;
    }

    template <class role>
    static typename bimap_type::template map_by<role>::type& by()
    {
        return Instance().template by<role>();
    }


}; // template class RelationDefinition


// } /* namespace yams */

#endif // YAMS__YAMS_H__INCLUDED

