/*
 * AnyEntity_test.cpp
 *
 *  Created on: 6 nov. 2012
 *      Author: Anthony Truchet <anthony.truchet@logilab.fr>
 */
#include "gtest/gtest.h"

#include <vector>

#include <boost/foreach.hpp>

#include "yams/utils.hh"

using std::cout;
using std::cerr;
using std::endl;
using std::vector;

// Convenience typedef
typedef vector<int>::iterator int_vect_it;
typedef vector<int>::const_iterator int_vect_const_it;
typedef ConvenientIterable<int_vect_it, int_vect_const_it> int_convenient_iterable;
typedef ConvenientIterable<int_vect_const_it> int_convenient_const_iterable;


TEST(ConvenientIteratorTest, BasicUse)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);
    // Creation
    // In C++'11 : auto i = iter(v);
    int_convenient_const_iterable i = iter(v);
    // Pair behaviour
    EXPECT_EQ(i.first, v.begin());
    EXPECT_EQ(i.second, v.end());
    EXPECT_EQ(i.begin(), i.first);
    EXPECT_EQ(i.end(), i.second);
    // Dereferencing
    EXPECT_EQ(*i, v[0]);
    // Pre-incrementation
    ++i;
    EXPECT_EQ(*i, v[1]);
    // Copy
    int_convenient_const_iterable i2(i);
    EXPECT_EQ(i, i2);
    // Post-incrementation
    EXPECT_EQ(i++, i2);
    EXPECT_EQ(*i, v[2]);
    // End of iteration
    EXPECT_TRUE(i);
    EXPECT_FALSE(++i);
}

TEST(ConvenientIteratorTest, BasicForLoop)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);

    int sum = 0;
    // NB long live auto in C++'11
    for (int_convenient_const_iterable i = iter(v) ; i; ++i)
    {
        sum += *i;
    }
    EXPECT_EQ(sum, 15);
}

TEST(ConvenientIteratorTest, BasicWhileLoop)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);

    int sum = 0;
    // NB long live auto in C++'11
    int_convenient_const_iterable i = iter(v);
    while (i)
    {
        int x = *i; ++i; // Do no touch i any further
        sum += x;
    }
    EXPECT_EQ(sum, 15);
}


TEST(ConvenientIteratorTest, BoostForEachLoop)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);
    int_convenient_const_iterable i = iter(v);
    int sum = 0;
    BOOST_FOREACH(int x, i)
    {
        sum += x;
    }
    EXPECT_EQ(sum, 15);
}

TEST(ConvenientIteratorTest, RefForLoop)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);

    // NB long live auto in C++'11
    for (int_convenient_iterable i = iter(v) ; i; ++i)
    {
        int& x = *i;
        x *= 2;
    }
    EXPECT_EQ(v[0], 6);
    EXPECT_EQ(v[1], 10);
    EXPECT_EQ(v[2], 14);
}

TEST(ConvenientIteratorTest, RefBoostForEachLoop)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);

    // NB long live auto in C++'11
    int_convenient_iterable i = iter(v) ;
    BOOST_FOREACH(int & x, i)
    {
        x *= 2;
    }
    EXPECT_EQ(v[0], 6);
    EXPECT_EQ(v[1], 10);
    EXPECT_EQ(v[2], 14);
}

TEST(ConvenientIteratorTest, AlgoFind)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);

    // NB long live auto in C++'11
    int_convenient_iterable i = iter(v) ;

    int_convenient_iterable::iterator found = i.find(5);
    EXPECT_EQ(found, v.begin() + 1);
}

template <class T> class Sum :
    public std::unary_function<const T&, T>
{
    T res;
public:
    Sum(T i = 0) : res(i) {}
    void operator()(const T& x) {res += x; }
    T result() const {return res;}
} ;

TEST(ConvenientIteratorTest, AlgoForEach)
{
    int vv[] = {3, 5, 7};
    vector<int> v(vv, vv + 3);

    // NB long live auto in C++'11
    int_convenient_iterable i = iter(v) ;

    Sum<double> sum;
    sum = i.for_each(sum);
    EXPECT_EQ(sum.result(), 15);
}
