/*
 * AnyEntity_test.cpp
 *
 *  Created on: 6 nov. 2012
 *      Author: Anthony Truchet <anthony.truchet@logilab.fr>
 */
#include "gtest/gtest.h"

#include <boost/foreach.hpp>

#include "yams/yams.hh"

using std::cout;
using std::cerr;
using std::endl;

struct Album;
struct Artist;

struct plays_in_rtype;

typedef RelationDefinition<plays_in_rtype, Artist, Album> plays_in_rdef;

struct Album  : public Entity<Album> {};
struct Artist : public Entity<Artist> {};

TEST(AnyEntityRelationTest, BasicPtrUse)
{
    Artist::pointer p_davis(new Artist());
    Album::const_pointer  p_kind_of_blue(new Album());

    Artist* raw_p_davis = p_davis.get();
    Artist::pointer p_davis2 = raw_p_davis->ptr();
    EXPECT_EQ(p_davis.use_count(), 2);
    EXPECT_EQ(p_davis2.use_count(), 2);
    EXPECT_EQ(p_davis, p_davis2);
    p_davis2.reset();
    EXPECT_EQ(p_davis.use_count(), 1);

    const Album* raw_p_kind_of_blue = p_kind_of_blue.get();
    Album::const_pointer p_kind_of_blue2 = raw_p_kind_of_blue->ptr();
    EXPECT_EQ(p_kind_of_blue.use_count(), 2);
    EXPECT_EQ(p_kind_of_blue2.use_count(), 2);
    EXPECT_EQ(p_kind_of_blue, p_kind_of_blue2);
    p_kind_of_blue2.reset();
    EXPECT_EQ(p_kind_of_blue.use_count(), 1);

}

TEST(AnyEntityRelationTest, BasicRelation)
{
    Artist::pointer p_davis(new Artist());
    Album::pointer  p_kind_of_blue(new Album());
    EXPECT_FALSE(plays_in_rdef::test(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::add(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::test(p_davis, p_kind_of_blue));
    EXPECT_FALSE(plays_in_rdef::add(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::test(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::remove(p_davis, p_kind_of_blue));
    EXPECT_FALSE(plays_in_rdef::test(p_davis, p_kind_of_blue));
    EXPECT_FALSE(plays_in_rdef::remove(p_davis, p_kind_of_blue));
    EXPECT_FALSE(plays_in_rdef::test(p_davis, p_kind_of_blue));
}

TEST(AnyEntityRelationTest, BasicRelationAltIface)
{
    Artist::pointer p_davis(new Artist());
    Album::pointer  p_kind_of_blue(new Album());
    EXPECT_FALSE(p_davis->rel<plays_in_rdef>(p_kind_of_blue));
    EXPECT_TRUE(p_davis->add<plays_in_rdef>(p_kind_of_blue));
    EXPECT_TRUE(p_davis->rel<plays_in_rdef>(p_kind_of_blue));
    EXPECT_FALSE(p_davis->add<plays_in_rdef>(p_kind_of_blue));
    EXPECT_TRUE(p_davis->rel<plays_in_rdef>(p_kind_of_blue));
    EXPECT_TRUE(p_davis->remove<plays_in_rdef>(p_kind_of_blue));
    EXPECT_FALSE(p_davis->rel<plays_in_rdef>(p_kind_of_blue));
    EXPECT_FALSE(p_davis->remove<plays_in_rdef>(p_kind_of_blue));
    EXPECT_FALSE(p_davis->rel<plays_in_rdef>(p_kind_of_blue));
}


TEST(AnyEntityRelationTest, RelationViews)
{
    Artist::pointer p_davis(new Artist());
    Artist::pointer p_chambers(new Artist());
    Album::pointer  p_kind_of_blue(new Album());
    // Album::pointer  p_blue_moods( new Album() );

    EXPECT_TRUE(plays_in_rdef::add(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::add(p_chambers, p_kind_of_blue));

    plays_in_rdef::subject_view::type& plays_in = plays_in_rdef::by<subject>();
    EXPECT_EQ(plays_in.count(p_davis), 1);
    BOOST_FOREACH(Album::pointer album, p_davis->iter<plays_in_rdef>())
    EXPECT_EQ(album, p_kind_of_blue);


    plays_in_rdef::object_view::type& features = plays_in_rdef::by<object>();
    EXPECT_EQ(features.count(p_kind_of_blue), 2);
    BOOST_FOREACH(Artist::pointer artist, p_kind_of_blue->rev_iter<plays_in_rdef>())
    EXPECT_TRUE((artist == p_davis)   or
                (artist == p_chambers));

    // Remove relations.
    EXPECT_TRUE(plays_in_rdef::remove(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::remove(p_chambers, p_kind_of_blue));

}


TEST(AnyEntityRelationTest, LoopOnSubject)
{
    // Artists.
    Artist::pointer p_davis(new Artist());
    Artist::pointer p_coltrane(new Artist());
    Artist::pointer p_chambers(new Artist());
    // Albums.
    Album::pointer  p_kind_of_blue(new Album());
    Album::pointer  p_tutu(new Album());
    Album::pointer  p_blue_moods( new Album() );
    Album::pointer  p_red_garland(new Album());

    // Add relations.
    EXPECT_TRUE(plays_in_rdef::add(p_davis, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::add(p_coltrane, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::add(p_chambers, p_kind_of_blue));
    EXPECT_TRUE(plays_in_rdef::add(p_davis, p_tutu));
    EXPECT_TRUE(plays_in_rdef::add(p_davis, p_blue_moods));
    EXPECT_TRUE(plays_in_rdef::add(p_chambers, p_red_garland));

    // View on 'Artist'.
    plays_in_rdef::subject_view::type& plays_in = plays_in_rdef::by<subject>();
    // A const iterator to walk trought on relations ('left view').
    typedef plays_in_rdef::bimap_type::map_by<subject>::const_iterator c_artist_iterator;
    c_artist_iterator it = plays_in.begin();

    // Loop which does not repeat the same key (i.e. subject). Suppose you have
    // same key which is contiguous in an std::unordered_multimap.
    // Here, you have 6 pairs of relation but just 3 different artists.
    unsigned int count = 0;
    for (; it != plays_in.end(); )
    {
        std::pair<c_artist_iterator, c_artist_iterator> p =
            plays_in.equal_range(it->get<subject>());
        it = p.second;
        ++count;
    }
    // XXX \todo replace this by a new feature/method in \c
    // RelationDefinition. Maybe you just can iterate on the bimap instance (a
    // protected singleton).
    EXPECT_EQ(count, 3);

    // XXX \todo remove all relations in order to clean future tests. Maybe add
    // this feature.
}

